#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment procedures

# (c) Issabella (tutorial) / Dinasset (filter) / 2021
#
#   History:
#
#   v1.0: 2021-09-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import random
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
FixWH=3000

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()
        
# here the single GEGL functions

def gegl_c2g(image, drawable, uradius, usamples, uiter):
    radius=int(uradius)      #dft 300   #ui_range (2, 1000)
    samples=int(usamples)    #dft 4     #ui_range (3, 17)
    iterations=int(uiter)    #dft 10    #ui_range (1, 30)
    enhance_shadows=False
    mystring="gegl:c2g radius=%d samples=%d iterations=%d enhance-shadows=%d"
    gegl_graph_string=mystring % (radius, samples, iterations, enhance_shadows)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_edge(image, drawable, ualgorithm): 
    algorithm="sobel" # 0=sobel,1=prewitt,2=gradient,3=roberts,4=differential,5=laplace
    amount=2.0        # value_range  (1.00 to 10.0)
    border=1          # 0=none,1=clamp,2=loop,3=black,4=white
    mystring="gegl:edge algorithm=%s amount=%f border=%d"
    gegl_graph_string=mystring % (algorithm, amount, border)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_median_blur (image, drawable, uradius):
    radius=float(uradius)
    percentile=50.0
    alpha_percentile=50.0
    gegl_graph_string="gegl:median-blur radius=%f percentile=%f alpha-percentile=%f" % (radius, percentile, alpha_percentile)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_gaussian_blur (image, drawable, ux, uy):    # gaussian blur  
    std_dev_x=float(ux)    #1.50     #user_range   (0.24, 100.0)
    std_dev_y=float(uy)    #1.50     #user_range   (0.24, 100.0)
    mystring="gegl:gaussian-blur std-dev-x=%f std-dev-y=%f"
    gegl_graph_string=mystring % (std_dev_x, std_dev_y)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_hue_chroma (image, drawable, uhue, uchroma, ulightness):    # Hue-Chroma  
    hue=uhue                #0.0   #value_range  (-180.0, 180.0) 
    chroma=uchroma          #12.0  #value_range  (-100.0, 100.0)  
    lightness=ulightness    #0.0   #value_range  (-100.0, 100.0)
    mystring="gegl:hue-chroma hue=%d chroma=%d lightness=%d"
    gegl_graph_string=mystring % (hue, chroma, lightness)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_saturation (image, drawable, uscale, ucolorspace):    # saturation 
    scale=uscale            #1.5  #user_range   (0.0, 2.0)
    colorspace=ucolorspace  #0    #value_range  (0, 1, 2)  Native, CIE-lab/lch, CIE-iuv
    mystring="gegl:saturation scale=%f colorspace=%s"
    gegl_graph_string=mystring % (scale, colorspace)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_exposure (image, drawable, ublack, uexpo):      
    black_level=ublack    # value_range (-0.1, 0.1)
    exposure=uexpo        # ui_range    (-10.0, 10.0)
    gegl_graph_string="gegl:exposure black-level=%f exposure=%f" % (black_level, exposure)
    gegl_graph(image, drawable,gegl_graph_string)
    
def gegl_unsharp_mask (image, drawable, usdev , uscale, uthreshold):      
    std_dev=usdev           #user_range   (0.0, 40.0)
    scale=uscale            #user_range   (0.0, 10.0)
    threshold=uthreshold    #user_range   (0.0, 1.0) 
    mystring="gegl:unsharp-mask std-dev=%f scale=%f threshold=%f"
    gegl_graph_string=mystring % (std_dev, scale, threshold)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_waterpixels (image, drawable, usize , usmooth, uregul, ufill):   
    size=usize                #user_range   (8, 256)
    smoothness=usmooth        #user_range   (0.0, 10.0)
    regularization=uregul     #user_range   (0, 50)
    fill=ufill                #average, random
    mystring="gegl:waterpixels size=%d smoothness=%f regularization=%d fill=%s"
    gegl_graph_string=mystring % (size, smoothness, regularization, fill)
    gegl_graph(image, drawable,gegl_graph_string)
    
############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    optiWidthHeight = FixWH       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    # adjust mean values
    adjustmeanv (newImage, newBG)
    # remove low values, remove high values (less) 
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                   1.00, 0.10, 0.95, False)
    newBG.opacity = 100.0
    
    return(newImage)

#---------------------------------------------------------------------------
# adjust values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer):
    # new version modified to take care of the existing mean in determining the theretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    # consider 3 classes
    if meanl < 85: thmean = (128+meanl)/2  #100
    elif meanl > 170: thmean = (127+meanl)/2  #155
    else: thmean = 127.5    # regular
    minmean = thmean-16
    maxmean = thmean+16
    
    modified=False
    if meanl < minmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        loop=0
        while meanl < minmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        loop=0
        while meanl > maxmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))
    
    return(inLayer)
     
#
# ================ *
#   M  A  I  N     *
# ---------------- *
   
def JvidEffectPA (inImage, inLayer, widthheight, texture, wpsize, arlequin,
                  doFlatten):
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
        
    OrigWH = inLayer.width+inLayer.height
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # display messages using error console
    
    newImage = createOutimage (inImage, inLayer)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width = Preprocessed.width
    height = Preprocessed.height
    
    if texture==True:
        # Apply Camouflage  #fx_camouflage 9,12,100,30,46,33,53,64,46,91,96,59,128,123,79
        pdb.plug_in_gmic_qt(newImage, Preprocessed, 1,2,
        "-v - -fx_camouflage 9,12,100,30,46,33,53,64,46,91,96,59,128,123,79 ")
        camouflage = newImage.layers[0]
        camouflage.mode = LAYER_MODE_HSV_VALUE
        camouflage.opacity = 15.0    
        camouflage.name = "[GMIC] -fx_camouflage 9,12,100,30,46,33,53,64,46,91,96,59,128,123,79"
        
        Preprocessed = pdb.gimp_image_merge_down (newImage, camouflage, CLIP_TO_IMAGE)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # prepare image for effects Blur + USM 
    # - Blur  
    BlurSharp = Preprocessed.copy()
    newImage.add_layer (BlurSharp, 0)
    uRadius=3.0
    BlurSharp.name = "Median Blurred 3.0" 
    gegl_median_blur (newImage, BlurSharp, uRadius)
    # - USM
    uSdev=3.0; uScale=1.0; uThreshold=0.0       # new parameters
    BlurSharp.name = "Median Blurred 3.0 then USM 3.0" 
    gegl_unsharp_mask (newImage, BlurSharp, uSdev, uScale, uThreshold)
    
    ToGray = BlurSharp.copy()
    newImage.add_layer (ToGray, 0)
    ToGray.name = "Colors to Gray 500,5,10"
    '''
    # GEGL Desaturate (C2G)
    RADIUS=1000; SAMPLES=15; ITER=10    # good but slow, use Gimp Desaturate instead, quicker
    gegl_c2g (newImage, ToGray, RADIUS, SAMPLES, ITER)
    '''
    # GIMP Desaturate 
    pdb.gimp_drawable_desaturate(ToGray, DESATURATE_VALUE)
    
    #pdb.gimp_display_new (newImage) 
    #return
     
    Waterpixels = ToGray.copy()
    newImage.add_layer (Waterpixels, 0)
    Waterpixels.name = "Waterpixels "+str(int(wpsize))+",1.0,0,average"
    Waterpixels.mode = LAYER_MODE_HSV_VALUE
    # GEGL Waterpixels
    SIZE=int(wpsize); SMOOTH=1.0; REGUL=0; FILL="average"
    gegl_waterpixels (newImage, Waterpixels, SIZE, SMOOTH, REGUL, FILL)
     
    WaterpixelsR = BlurSharp.copy()
    newImage.add_layer (WaterpixelsR, 0)
    WaterpixelsR.name = "Waterpixels Random (size*3),1.0,0,random"
    WaterpixelsR.mode = LAYER_MODE_SOFTLIGHT
    WaterpixelsR.opacity = arlequin*1.0
    # GEGL Waterpixels
    SIZE=min(255,int(wpsize*3)); SMOOTH=1.0; REGUL=0; FILL="random"
    gegl_waterpixels (newImage, WaterpixelsR, SIZE, SMOOTH, REGUL, FILL)
    
    pdb.gimp_item_set_visible (ToGray, False)
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # prepare first base  
    Base1 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base1, 0)
    Base1.name = "Base for Graphic Boost "
    Base1.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Effect: Graphic Boost
    pdb.plug_in_gmic_qt(newImage, Base1, 1,2 ,
    #[G'MIC] Graphic Boost: -fx_graphic_boost4 1.25,2,0,0.15,14,0,1,0.5,0.45,2,0,0,0,1,1,1,0.5,0.45,1
    "-v - -fx_graphic_boost4 1.25,1.5,0,0.15,14,0,1,0.5,0.45,2,0,0,0,1,1,1,0.5,0.45,1 ") # new:1.5
    GBoost_Layer = newImage.layers[0]
    GBoost_Layer.mode = LAYER_MODE_HSV_VALUE
    GBoost_Layer.opacity = 65.0 #100.0  
    GBoost_Layer.name = "[GMIC] -fx_graphic_boost4 (default values)"
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # combine with original
    OriginalOnTop = Original.copy()
    newImage.add_layer (OriginalOnTop, 0)
    pdb.gimp_item_set_visible (OriginalOnTop, True)
    OriginalOnTop.opacity = 35.0
    OriginalOnTop.name = "Original on Top for merging colours and structure"
    
    # prepare second base  
    Base2 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base2, 0)
    Base2.name = "Base for Strokes "
    Base2.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Effect: Strokes by Hair Locks  gtutor_hairlock 0.5,0.5,0.5,255,255,0,255,0.5,45,0.5,0.5,0
    pdb.plug_in_gmic_qt(newImage, Base2, 1,2,
    #[G'MIC] Hair Locks: -gtutor_hairlock 0.5,0.5,0.5,255,255,255,255,0.5,45,0.5,0.5,0 std
    "-v - -gtutor_hairlock 0.55,0.55,0.55,255,255,64,255,0.5,45,0.5,0.5,0 ")
    HairLocks_Layer = newImage.layers[0]
    HairLocks_Layer.mode = LAYER_MODE_LINEAR_BURN
    HairLocks_Layer.opacity = 80.0  #100.0  
    HairLocks_Layer.name = "[GMIC] -gtutor_hairlock (yellowish)"
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # prepare third base  
    Base3 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base3, 0)
    Base3.name = "Base for Canvas "
    Base3.mode = LAYER_MODE_NORMAL
    adjustmeanv (newImage, Base3)
    # remove low values
    #pdb.gimp_drawable_levels (Base3, HISTOGRAM_VALUE, 0.0, 1.0, False,
    #                               1.00, 0.10, 0.99, False)
    pdb.gimp_drawable_curves_spline (Base3,HISTOGRAM_VALUE,8,
                                     (0.0,0.01,0.04,0.12,0.5,0.5,1.0,1.0))
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Apply Canvas  
    pdb.plug_in_gmic_qt(newImage, Base3, 1,0,
    "-v - -texturize_canvas 20,4,0.8 ")
    
    #pdb.gimp_display_new (newImage)
    #return
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "outcome"
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with JVID effect P (Fake Acrylic Paint Effect) Arlequin @"
    outcome.name = outcome.name+str(int(arlequin))+" using size "+str(int(wpsize)) 
    
    if widthheight==0: widthheight=OrigWH
    uRatio = widthheight*1.00/FixWH*1.00
    uWidth = int(width*uRatio)
    uHeight = int(height*uRatio)
    #info_msg ("SCALING")   # not during tests, scaling is slow
    pdb.gimp_image_scale (newImage, uWidth, uHeight)   # to user chosen size
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="issabella/dinasset"
menu="<Image>/Jvid/"
copyrightYear="2021"
desc="JVID Effect PA (Fake Acrylic Paint Effect) Arlequin"

register(
    "JvidEffectPA",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 6000, 100)),
        (PF_TOGGLE, "texture", "Texturize Image?", False),
        (PF_ADJUSTMENT, "wpsize", "Waterpixel size", 24, (8, 128, 4)),
        (PF_ADJUSTMENT, "arlequin", "Arlequin colours percent", 25, (0, 100, 1)),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    JvidEffectPA,
    menu=menu
)

main()
